from typing import Any

from scy import exceptions
from scy.tokens import TokenType, Token, KEYWORDS
from scy.utils import find_line


class Tokenizer:
    source: str
    filename: str
    tokens: list[Token]
    start: int
    current: int
    line: int
    start_column: int
    column: int

    def __init__(self, source: str, filename: str = '<unknown>') -> None:
        self.source = source
        self.filename = filename
        self.tokens = []
        self.start = 0
        self.current = 0
        self.line = 1
        self.start_column = 0
        self.column = 0

    def tokenize(self) -> list[Token]:
        while not self.is_at_end():
            self.start = self.current
            self.start_column = self.column
            self.scan()

        self.tokens.append(Token(TokenType.EOF, '', self.line, self.start_column, self.current))
        return self.tokens

    def error(self, text: str) -> SyntaxError:
        return SyntaxError(text, (self.filename, self.line, self.start_column + 1,
                           find_line(self.source, self.current)))

    def scan(self) -> None:
        c: str = self.advance()
        if c == '(': # I can't wait for Python 3.10 to have match..case lol
            self.add_token(TokenType.LEFT_PAREN)
        elif c == ')':
            self.add_token(TokenType.RIGHT_PAREN)
        elif c == '{':
            self.add_token(TokenType.LEFT_BRACE)
        elif c == '}':
            self.add_token(TokenType.RIGHT_BRACE)
        elif c == ',':
            self.add_token(TokenType.COMMA)
        elif c == '.':
            self.add_token(TokenType.DOT)
        elif c == '-':
            self.add_token(TokenType.MINUS)
        elif c == '+':
            self.add_token(TokenType.PLUS)
        elif c == '~':
            self.add_token(TokenType.TILDE)
        elif c == ';':
            self.add_token(TokenType.SEMICOLON)
        elif c == '%':
            self.add_token(TokenType.PERCENT)
        elif c == '^':
            self.add_token(TokenType.CARET)
        elif c == ':':
            self.add_token(TokenType.COLON)
        elif c == '@':
            self.add_token(TokenType.AT)
        elif c == '*':
            self.add_token(TokenType.STAR_STAR if self.match_('*') else TokenType.STAR)
        elif c == '!':
            self.add_token(TokenType.BANG_EQUAL if self.match_('=') else TokenType.BANG)
        elif c == '=':
            self.add_token(TokenType.EQUAL_EQUAL if self.match_('=') else TokenType.EQUAL)
        elif c == '<':
            if self.match_('='):
                tok = TokenType.LESS_EQUAL
            elif self.match_('<'):
                tok = TokenType.LESS_LESS
            else:
                tok = TokenType.LESS
            self.add_token(tok)
        elif c == '>':
            if self.match_('='):
                tok = TokenType.GREATER_EQUAL
            elif self.match_('>'):
                tok = TokenType.GREATER_GREATER
            else:
                tok = TokenType.GREATER
            self.add_token(tok)
        elif c == '/':
            self.add_token(TokenType.SLASH_SLASH if self.match_('/') else TokenType.SLASH)
        elif c == '&':
            self.add_token(TokenType.AMPERSAND_AMPERSAND if self.match_('&') else TokenType.AMPERSAND)
        elif c == '|':
            self.add_token(TokenType.PIPE_PIPE if self.match_('|') else TokenType.PIPE)
        elif c == '#':
            while self.peek() != '\n' and not self.is_at_end():
                self.advance()

        elif c in ' \r\t':
            pass

        elif c == '\n':
            self.line += 1
            self.column = 0

        elif c in '"\'':
            self.string(c)

        elif self.is_digit(c):
            self.number()
        elif self.is_alpha(c):
            self.identifier()

        else:
            raise self.error(exceptions.UNEXPECTED_CHARACTER)

    def identifier(self) -> None:
        while self.is_alpha_numeric(self.peek()):
            self.advance()
        text: str = self.source[self.start:self.current]
        type: TokenType = KEYWORDS.get(text, TokenType.IDENTIFIER)
        if type is None:
            raise self.error(exceptions.RESERVED_KEYWORD % text)
        self.add_token(type)

    def number(self) -> None:
        while self.is_digit(self.peek()):
            self.advance()
        if self.peek() == '.' and self.is_digit(self.peek_next()):
            self.advance()
            while self.is_digit(self.peek()):
                self.advance()
            return self.add_token_literal(TokenType.DECIMAL, float(self.source[self.start:self.current]))
        self.add_token_literal(TokenType.INTEGER, int(self.source[self.start:self.current]))

    def escape(self) -> str:
        self.advance()
        code = self.peek()
        if code == 'n':
            return '\n'
        elif code == 't':
            return '\t'
        elif code == '"':
            return '"'
        elif code == "'":
            return "'"
        elif code == 'x':
            self.advance()
            data = self.advance() + self.peek()
            try:
                return chr(int(data, 16))
            except ValueError:
                raise self.error(exceptions.INVALID_HEX_STRING % data) from None
        elif code == 'u':
            self.advance()
            data = self.advance()
            for _ in range(2):
                data += self.advance()
            data += self.peek()
            try:
                return chr(int(data, 16))
            except ValueError:
                raise self.error(exceptions.INVALID_HEX_STRING % data) from None
        elif code == 'U':
            self.advance()
            data = self.advance()
            for _ in range(6):
                data += self.advance()
            data += self.peek()
            try:
                return chr(int(data, 16))
            except ValueError:
                raise self.error(exceptions.INVALID_HEX_STRING % data) from None

    def string(self, end: str) -> None:
        result = ''
        while self.peek() != end and not self.is_at_end():
            if self.peek() == '\n':
                raise self.error(exceptions.MULTILINE_STRINGS_NOT_SUPPORTED)
            elif self.peek() == '\\':
                result += self.escape()
            else:
                result += self.peek()
            self.advance()
        if self.is_at_end():
            raise self.error(exceptions.EOF_DURING_STRING)
        self.advance()
        self.add_token_literal(TokenType.STRING, result)

    def match_(self, expected: str) -> bool:
        if self.is_at_end():
            return False
        if self.source[self.current] != expected:
            return False

        self.current += 1
        self.column += 1
        return True

    def peek(self) -> str:
        if self.is_at_end():
            return '\0'
        return self.source[self.current]

    def peek_next(self) -> str:
        if self.current + 1 >= len(self.source):
            return '\0'
        return self.source[self.current + 1]

    def is_alpha(self, c: str) -> bool:
        return ((c >= 'a' and c <= 'z')
             or (c >= 'A' and c <= 'Z')
             or c == '_')

    def is_alpha_numeric(self, c: str) -> bool:
        return self.is_alpha(c) or self.is_digit(c)

    def is_digit(self, c: str) -> bool:
        return c >= '0' and c <= '9'

    def is_at_end(self) -> bool:
        return self.current >= len(self.source)

    def advance(self) -> str:
        if self.is_at_end():
            raise self.error(exceptions.UNEXPECTED_EOF)
        char = self.source[self.current]
        self.current += 1
        self.column += 1
        return char

    def add_token(self, type: TokenType) -> None:
        self.add_token_literal(type, None)

    def add_token_literal(self, type: TokenType, literal: Any) -> None:
        text: str = self.source[self.start:self.current]
        self.tokens.append(Token(
            type,
            text,
            self.line,
            self.start_column,
            self.current,
            literal
        ))


def tokenize(source: str, filename: str = '<unknown>') -> list[Token]:
    tokenizer = Tokenizer(source, filename)
    return tokenizer.tokenize()



# Tokenizer exceptions
UNEXPECTED_CHARACTER = 'Unexpected character.'
RESERVED_KEYWORD = 'Reserved keyword %r.'
INVALID_HEX_STRING = 'Invalid hex string %r.'
MULTILINE_STRINGS_NOT_SUPPORTED = 'Multiline strings not supported.'
EOF_DURING_STRING = 'EOF reached during string.'
UNEXPECTED_EOF = 'Unexpected EOF.'

# Parser exceptions
INVALID_ASSIGNMENT = 'Invalid assignment target.'
ITERATION_INVALID_ASSIGNMENT = 'Invalid assignment target for iteration.'
EXPECT_EXPRESSOIN = 'Expect expression.'
INVALID_ASYNC = 'Async keyword not supported with %r statements.'
INVALID_ASYNC_EXPR = 'Async keyword not supported with expression statements.'
INVALID_ASYNC_FOR = "Async for loops only compatible with iteration (using ':' syntax)."
EXPECT_PROPERTY_NAME = "Expect property name after '.'."



from types import CodeType
from typing import Any, Mapping, Optional, Union
import os

from scy.backend import parse


def scy_compile(
    source: str,
    filename: Union[str, os.PathLike],
    mode: str,
    flags: int = 0,
    dont_inherit: int = False,
    optimize: int = -1) -> CodeType:
    filename = os.fspath(filename)
    tree = parse(source, filename, mode)
    return compile(tree, filename, mode, flags, dont_inherit, optimize)


def scy_eval(
    expression: Union[str, CodeType],
    globals: Optional[dict[str, Any]] = None,
    locals: Optional[Mapping[str, Any]] = None) -> Any:
    if not isinstance(expression, CodeType):
        expression = scy_compile(expression, '<string>', 'eval')
    return eval(expression, globals, locals)


def scy_exec(
    expression: Union[str, CodeType],
    globals: Optional[dict[str, Any]] = None,
    locals: Optional[Mapping[str, Any]] = None) -> Any:
    if not isinstance(expression, CodeType):
        expression = scy_compile(expression, '<string>', 'exec')
    return exec(expression, globals, locals)



import ast


class NodeCounter(ast.NodeVisitor):
    count: int

    def __init__(self) -> None:
        super().__init__()
        self.count = 0

    def visit(self, node: ast.AST) -> None:
        self.generic_visit(node)
        self.count += 1


def find_line(code: str, index: int) -> str:
    start = code.rfind('\n', 0, index) + 1
    code = code[start:]
    end = code.find('\n')
    if end == -1:
        end = len(code)
    print(repr(code[:end]))
    return code[:end]


def count_nodes(tree: ast.AST) -> int:
    counter = NodeCounter()
    counter.visit(tree)
    return counter.count



import ast
from scy import exceptions
from typing import Any, Union
from scy.tokenizer import Tokenizer
from scy.utils import find_line

from scy.tokens import BINARY_OPERATORS, COMPARISON_OPERATORS, TokenGroup, TokenType, Token, UNARY_OPERATORS


ASSIGNABLES = (
    ast.Attribute,
    ast.Subscript,
    ast.Starred,
    ast.Name,
    ast.List,
    ast.Tuple,
)


class Parser:
    tokens: list[Token]
    filename: str
    source: str
    current: int

    def __init__(self, tokens: list[Token], filename: str, source: str) -> None:
        self.tokens = tokens
        self.filename = filename
        self.source = source
        self.current = 0

    def declaration(self) -> list[ast.stmt]:
        is_async = self.match_(TokenType.ASYNC)
        if self.match_(TokenType.DEF):
            return [self.function(self.peek(), is_async)]
        elif self.match_(TokenType.CLASS):
            self.raise_if_async(is_async)
            return [self.class_(self.peek())]
        return self.statement(is_async)

    def function(self, creator: Token, is_async: bool) -> ast.FunctionDef:
        klass = ast.AsyncFunctionDef if is_async else ast.FunctionDef
        name = self.consume(TokenType.IDENTIFIER, f'Expect function name.')
        self.consume(TokenType.LEFT_PAREN, f"Expect '(' after function name.")
        arguments = self.parse_args_def()
        self.consume(TokenType.LEFT_BRACE, f"Expect '{{' before function body.")
        body = self.block()
        if not body:
            body = [self.ast_token(klass=ast.Pass)]
        return self.ast_token(name.lexeme, arguments, body, [],
            klass=klass, first=creator, last=self.previous())

    def class_(self, creator: Token) -> ast.FunctionDef:
        name = self.consume(TokenType.IDENTIFIER, f'Expect class name.')
        if self.match_(TokenType.LEFT_PAREN):
            args, kwargs, paren = self.parse_args_call()
        else:
            args, kwargs, paren = [], [], None
        self.consume(TokenType.LEFT_BRACE, f"Expect '{{' before class body.")
        body = self.block()
        if not body:
            body = [self.ast_token(klass=ast.Pass)]
        return self.ast_token(name.lexeme, args, kwargs, body, [],
            klass=ast.ClassDef, first=creator, last=self.previous())

    def parse_args_def(self) -> ast.arguments:
        arguments = ast.arguments([], [], None, [], [], None, [])
        if not self.check(TokenType.RIGHT_PAREN):
            while True:
                self.consume(TokenType.IDENTIFIER, 'Expect argument name.')
                arguments.args.append(self.ast_token(self.previous().lexeme, klass=ast.arg))
                if not self.match_(TokenType.COMMA):
                    break
        self.consume(TokenType.RIGHT_PAREN, "Expect ')' after arguments.")
        return arguments

    def raise_if_async(self, is_async: bool) -> None:
        if is_async:
            raise self.error(self.previous(), exceptions.INVALID_ASYNC % self.previous().lexeme)

    def statement(self, is_async: bool = False) -> list[ast.stmt]:
        if self.match_(TokenType.FOR):
            return self.for_statement(is_async)
        elif self.match_(TokenType.IF):
            self.raise_if_async(is_async)
            return [self.if_statement()]
        elif self.match_(TokenType.RETURN):
            self.raise_if_async(is_async)
            return [self.return_statement()]
        elif self.match_(TokenType.WHILE):
            self.raise_if_async(is_async)
            return [self.while_statement()]
        elif self.match_(TokenType.BREAK, TokenType.CONTINUE):
            self.raise_if_async(is_async)
            word = self.previous()
            self.consume(TokenType.SEMICOLON, f"Expect ';' after {word.lexeme}.")
            return [self.ast_token(klass={
                TokenType.BREAK:    ast.Break,
                TokenType.CONTINUE: ast.Continue,
            }.get(word.type), first=word)]
        if is_async:
            raise self.error(self.peek(), exceptions.INVALID_ASYNC_EXPR)
        return [self.expression_statement()]

    def return_statement(self) -> ast.stmt:
        keyword = self.previous()
        if self.match_(TokenType.SEMICOLON):
            return self.ast_token(klass=ast.Return, first=keyword)
        else:
            value = self.expression(False)
            last = self.previous()
            self.consume(TokenType.SEMICOLON, "Expect ';' after return value.")
            return self.ast_token(value, klass=ast.Return, first=keyword, last=last)

    def for_statement(self, is_async: bool) -> list[ast.stmt]:
        for_word = self.previous()
        self.consume(TokenType.LEFT_PAREN, "Expect '(' after 'for'.")
        if self.match_(TokenType.SEMICOLON):
            initializer = None
        else:
            initializer = self.expression_statement(
                (TokenType.SEMICOLON, TokenType.COLON),
                "Expect ';' or ':' after statement."
            )
            if self.previous().type == TokenType.COLON:
                if isinstance(initializer.value, ast.Name):
                    initializer.value.ctx = ast.Store()
                    return [self.for_in_statement(initializer.value, for_word, is_async)]
                else:
                    raise self.error(self.previous(), exceptions.ITERATION_INVALID_ASSIGNMENT)
            if is_async:
                raise self.error(self.peek(), exceptions.INVALID_ASYNC_FOR)
        if self.check(TokenType.SEMICOLON):
            condition = None
            condition_tok = self.peek()
        else:
            condition = self.expression(False)
            condition_tok = None
        self.consume(TokenType.SEMICOLON, "Expect ';' after loop condition.")
        if self.match_(TokenType.RIGHT_PAREN):
            increment = None
        else:
            increment = self.expression_statement(TokenType.RIGHT_PAREN,
                                                  "Expect ')' after for clauses")
        body = self.optional_block(increment is None)
        if increment is not None:
            body.append(increment)
        if condition is None:
            condition = self.ast_token(True, first=condition_tok)
        if self.match_(TokenType.ELSE):
            else_branch = self.optional_block()
        else:
            else_branch = []
        result = [self.ast_token(condition, body, else_branch,
                                 klass=ast.While, first=for_word, last=self.previous())]
        if initializer is not None:
            result.insert(0, initializer)
        return result

    def for_in_statement(self, target: ast.Name, for_word: Token, is_async: bool) -> Union[ast.For, ast.AsyncFor]:
        klass = ast.AsyncFor if is_async else ast.For
        iterable = self.expression(False)
        self.consume(TokenType.RIGHT_PAREN, "Expect ')' after for clauses")
        body = self.optional_block()
        if self.match_(TokenType.ELSE):
            else_branch = self.optional_block()
        else:
            else_branch = []
        return self.ast_token(target, iterable, body, else_branch,
                              klass=klass, first=for_word, last=self.previous())

    def if_statement(self) -> ast.If:
        if_word = self.previous()
        self.consume(TokenType.LEFT_PAREN, "Expect '(' after 'if'.")
        condition = self.expression(False)
        self.consume(TokenType.RIGHT_PAREN, "Expect ')' after if condition.")
        then_branch = self.optional_block()
        if self.match_(TokenType.ELSE):
            else_branch = self.optional_block()
        else:
            else_branch = []
        return self.ast_token(condition, then_branch, else_branch,
                              klass=ast.If, first=if_word, last=self.previous())

    def while_statement(self) -> ast.stmt:
        while_word = self.previous()
        self.consume(TokenType.LEFT_PAREN, "Expect '(' after 'while'.")
        condition = self.expression(False)
        self.consume(TokenType.RIGHT_PAREN, "Expect ')' after condition")
        body = self.optional_block()
        if self.match_(TokenType.ELSE):
            else_branch = self.optional_block()
        else:
            else_branch = []
        return self.ast_token(condition, body, else_branch,
                              klass=ast.While, first=while_word, last=self.previous())

    def expression_statement(self, end: Union[TokenType, tuple[TokenType]] = TokenType.SEMICOLON,
                                   error: str = "Expect ';' after statement.") -> Union[ast.Expr]:
        expr = self.expression()
        if self.match_(TokenType.EQUAL):
            if not isinstance(expr, ASSIGNABLES):
                raise self.error(self.previous(), exceptions.INVALID_ASSIGNMENT)
            extra = [expr, self.expression()]
            extra[0].ctx = ast.Store()
            while self.match_(TokenType.EQUAL):
                if isinstance(extra[-1], ASSIGNABLES):
                    extra[-1].ctx = ast.Store()
                else:
                    raise self.error(self.previous(), exceptions.INVALID_ASSIGNMENT)
                extra.append(self.expression())
            value = extra.pop()
            statement = ast.Assign(targets=extra, value=value, **self.get_loc(extra[0], value))
        else:
            statement = ast.Expr(expr, **self.get_loc(expr, expr))
        if isinstance(end, tuple):
            self.consume_any(end, error)
        else:
            self.consume(end, error)
        return statement

    def optional_block(self, fill_empty: bool = True) -> list[ast.stmt]:
        if self.match_(TokenType.LEFT_BRACE):
            result = self.block()
            if not result and fill_empty:
                result = [self.ast_token(klass=ast.Pass)]
            return result
        elif self.match_(TokenType.SEMICOLON):
            # This is done so that ast.unparse still generates valid syntax for empty blocks
            if fill_empty:
                return [self.ast_token(klass=ast.Pass)]
            return []
        return self.declaration()

    def block(self) -> list[ast.stmt]:
        statements = []
        while not self.check(TokenType.RIGHT_BRACE) and not self.is_at_end():
            statements.extend(self.declaration())
        self.consume(TokenType.RIGHT_BRACE, "Expect '}' after block.")
        return statements

    def expression(self, toplevel: bool = True) -> ast.expr:
        if not toplevel:
            return self.assignment_expression()
        return self.yield_()

    def assignment_expression(self) -> ast.expr:
        expr = self.yield_()
        if self.match_(TokenType.EQUAL):
            equals = self.previous()
            value = self.assignment_expression()
            if isinstance(expr, ASSIGNABLES):
                expr.ctx = ast.Store()
                return ast.NamedExpr(expr, value, **self.get_loc(expr, value))
            raise self.error(equals, exceptions.INVALID_ASSIGNMENT)
        return expr

    def yield_(self) -> ast.expr:
        if self.match_(TokenType.YIELD):
            first_word = self.previous()
            if self.match_(TokenType.FROM):
                klass = ast.YieldFrom
                second_word = self.previous()
            else:
                klass = ast.Yield
                second_word = first_word
            try:
                value = self.or_()
            except SyntaxError as e:
                if e.msg == exceptions.EXPECT_EXPRESSOIN:
                    return self.ast_token(klass=klass, first=first_word, last=second_word)
                else:
                    raise
            return self.ast_token(value, klass=klass, first=first_word, last=self.previous())
        return self.or_()

    def or_(self) -> ast.expr:
        left = self.and_()
        if self.match_(TokenType.PIPE_PIPE):
            values = [left, self.and_()]
            while self.match_(TokenType.PIPE_PIPE):
                values.append(self.and_())
            return ast.BoolOp(ast.Or(), values, **self.get_loc(left, values[-1]))
        return left

    def and_(self) -> ast.expr:
        left = self.not_()
        if self.match_(TokenType.AMPERSAND_AMPERSAND):
            values = [left, self.not_()]
            while self.match_(TokenType.AMPERSAND_AMPERSAND):
                values.append(self.not_())
            return ast.BoolOp(ast.And(), values, **self.get_loc(left, values[-1]))
        return left

    def not_(self) -> ast.expr:
        if self.match_(TokenType.BANG):
            right = self.comparison()
            return ast.UnaryOp(ast.Not(), right, **self.get_loc(right, right))
        return self.comparison()

    def comparison(self) -> ast.expr:
        left: ast.expr = self.bit_or()
        operators: list[ast.cmpop] = []
        extra: list[ast.expr] = []
        while self.match_(*TokenGroup.SINGLE_COMPARISON, TokenType.NOT):
            if self.previous().type == TokenType.IS:
                if self.match_(TokenType.NOT):
                    operator = ast.IsNot()
                else:
                    operator = ast.Is()
            elif self.previous().type == TokenType.NOT:
                self.consume(TokenType.IN, "'in' must follow 'not' in comparison.")
                operator = ast.NotIn()
            else:
                operator = COMPARISON_OPERATORS[self.previous().type]()
            right = self.bit_or()
            operators.append(operator)
            extra.append(right)
        if operators:
            return ast.Compare(left, operators, extra, **self.get_loc(left, extra[-1]))
        else:
            return left

    def bit_or(self):
        left = self.bit_xor()
        while self.match_(TokenType.PIPE):
            right = self.bit_xor()
            left = ast.BinOp(left, ast.BitOr(), right, **self.get_loc(left, right))
        return left

    def bit_xor(self):
        left = self.bit_and()
        while self.match_(TokenType.CARET):
            right = self.bit_and()
            left = ast.BinOp(left, ast.BitXor(), right, **self.get_loc(left, right))
        return left

    def bit_and(self):
        left = self.bit_shift()
        while self.match_(TokenType.AMPERSAND):
            right = self.bit_shift()
            left = ast.BinOp(left, ast.BitAnd(), right, **self.get_loc(left, right))
        return left

    def bit_shift(self):
        left = self.term()
        while self.match_(*TokenGroup.BIT_SHIFT):
            operator = BINARY_OPERATORS[self.previous().type]()
            right = self.term()
            left = ast.BinOp(left, operator, right, **self.get_loc(left, right))
        return left

    def term(self) -> ast.expr:
        left = self.factor()
        while self.match_(*TokenGroup.TERMS):
            operator = BINARY_OPERATORS[self.previous().type]()
            right = self.factor()
            left = ast.BinOp(left, operator, right, **self.get_loc(left, right))
        return left

    def factor(self) -> ast.expr:
        left = self.unary()
        while self.match_(*TokenGroup.FACTORS):
            operator = BINARY_OPERATORS[self.previous().type]()
            right = self.unary()
            left = ast.BinOp(left, operator, right, **self.get_loc(left, right))
        return left

    def unary(self) -> ast.expr:
        if self.match_(*TokenGroup.UNARY_LOW):
            operator = UNARY_OPERATORS[self.previous().type]()
            right = self.unary()
            return ast.UnaryOp(operator, right, **self.get_loc(right, right))
        return self.power()

    def power(self) -> ast.expr:
        left = self.await_()
        while self.match_(TokenType.STAR_STAR):
            right = self.await_()
            left = ast.BinOp(left, ast.Pow(), right, **self.get_loc(left, right))
        return left

    def await_(self) -> ast.expr:
        if self.match_(TokenType.AWAIT):
            first_word = self.previous()
            value = self.call()
            return self.ast_token(value, klass=ast.Await, first=first_word, last=self.previous())
        return self.call()

    def call(self) -> ast.expr:
        expr = self.primary()
        while True:
            if self.match_(TokenType.LEFT_PAREN):
                expr = self.finish_call(expr)
            elif self.match_(TokenType.DOT):
                name = self.consume(TokenType.IDENTIFIER, exceptions.EXPECT_PROPERTY_NAME)
                expr = ast.Attribute(expr, name.lexeme, ast.Load(),
                    lineno=expr.lineno, end_lineno=name.line,
                    col_offset=expr.col_offset, end_col_offset=name.column + len(name.lexeme)
                )
            else:
                break
        return expr

    def finish_call(self, callee: ast.expr) -> ast.expr:
        args, kwargs, paren = self.parse_args_call()
        return ast.Call(callee, args, kwargs,
            lineno=callee.lineno, end_lineno=paren.line,
            col_offset=callee.col_offset, end_col_offset=paren.column + 1
        )

    def parse_args_call(self) -> tuple[list[ast.expr], list[ast.keyword], Token]:
        args = []
        kwargs = []
        if not self.check(TokenType.RIGHT_PAREN):
            while True:
                args.append(self.expression(False))
                if not self.match_(TokenType.COMMA):
                    break
        paren = self.consume(TokenType.RIGHT_PAREN, "Expect ')' after arguments")
        return args, kwargs, paren

    def primary(self) -> ast.expr:
        if self.match_(TokenType.FALSE):
            return self.ast_token(False)
        elif self.match_(TokenType.TRUE):
            return self.ast_token(True)
        elif self.match_(TokenType.NONE):
            return self.ast_token(None)
        elif self.match_(*TokenGroup.LITERALS):
            return self.ast_token(self.previous().literal)
        elif self.match_(TokenType.IDENTIFIER):
            tok = self.previous()
            return self.ast_token(tok.lexeme, ast.Load(), klass=ast.Name)
        elif self.match_(TokenType.LEFT_PAREN):
            expr = self.expression(False)
            self.consume(TokenType.RIGHT_PAREN, "Expect ')' after expression.")
            return expr
        else:
            raise self.error(self.peek(), exceptions.EXPECT_EXPRESSOIN)

    def ast_token(self, *args, klass: type[ast.AST] = ast.Constant,
                  first: Token = None, last: Token = None) -> Any:
        if first is None:
            first = self.previous()
        if last is None:
            last = first
        result = klass(*args)
        result.lineno = first.line
        result.end_lineno = last.line
        result.col_offset = first.column
        result.end_col_offset = last.column + len(last.lexeme)
        return result

    def get_loc(self, left: ast.AST, right: ast.AST):
        return dict(
            lineno=left.lineno, end_lineno=right.end_lineno,
            col_offset=left.col_offset, end_col_offset=right.end_col_offset
        )

    def match_(self, *types: TokenType) -> bool:
        if any(self.check(type) for type in types):
            self.advance()
            return True
        return False

    def consume(self, type: TokenType, message: str) -> Token:
        if self.check(type):
            return self.advance()
        raise self.error(self.peek(), message)

    def consume_any(self, types: tuple[TokenType], message: str) -> Token:
        if self.match_(*types):
            return self.peek()
        raise self.error(self.peek(), message)

    def error(self, token: Token, message: str) -> SyntaxError:
        return SyntaxError(message, (self.filename, token.line,
                                     token.column + 1,
                           find_line(self.source, token.index)))

    def check(self, type: TokenType) -> bool:
        if self.is_at_end():
            return False
        return self.peek().type == type

    def advance(self) -> Token:
        if not self.is_at_end():
            self.current += 1
        return self.previous()

    def is_at_end(self) -> bool:
        return self.peek().type == TokenType.EOF

    def peek(self) -> Token:
        return self.tokens[self.current]

    def previous(self) -> Token:
        return self.tokens[self.current - 1]

    def parse(self, mode: str = 'exec') -> Union[ast.Expression, ast.Module]:
        if mode == 'eval':
            return ast.Expression(body=self.expression())
        elif mode == 'exec':
            statements = []
            while not self.is_at_end():
                statements.extend(self.declaration())
            return ast.Module(body=statements, type_ignores=[])
        raise ValueError(f'No such parse mode named {mode!r}')


def parse_tree(tokens: list[Token], mode: str = 'exec', filename: str = '<unknown>', source: str = '') -> Union[ast.Expression, ast.Module]:
    parser: Parser = Parser(tokens, filename, source)
    return parser.parse(mode)



import ast
from typing import Union

from scy.parser import parse_tree
from scy.tokens import Token
from scy.tokenizer import tokenize


def parse(source, filename: str = '<unknown>', mode: str = 'exec') -> Union[ast.Expression, ast.Module]:
    tokens: list[Token] = tokenize(source, filename)
    tree = parse_tree(tokens, mode, filename, source)
    return tree



import ast
import argparse
import builtins
import sys
from typing import Optional
import time

from scy.backend import parse
from scy.utils import count_nodes

parser = argparse.ArgumentParser('python -m scy')
parser.add_argument('script', type=argparse.FileType('r'))
parser.add_argument('-M', '--mode', choices=['auto', 'run', 'dump', 'py', 'compile_only'], default='auto')


# This was copied from the runpy module
# It's copied here just in case it's removed from runpy
def _run_code(code, run_globals, init_globals=None,
              mod_name=None, mod_spec=None,
              pkg_name=None, script_name=None):
    """Helper to run code in nominated namespace"""
    if init_globals is not None:
        run_globals.update(init_globals)
    if mod_spec is None:
        loader = None
        fname = script_name
        cached = None
    else:
        loader = mod_spec.loader
        fname = mod_spec.origin
        cached = mod_spec.cached
        if pkg_name is None:
            pkg_name = mod_spec.parent
    run_globals.update(__name__ = mod_name,
                       __file__ = fname,
                       __cached__ = cached,
                       __doc__ = None,
                       __loader__ = loader,
                       __package__ = pkg_name,
                       __spec__ = mod_spec)
    exec(code, run_globals)
    return run_globals


def main() -> int:
    args = parser.parse_args()
    if args.mode == 'auto':
        args.mode = 'run'
    source = args.script.read()
    try:
        filename = args.script.name
    except Exception:
        filename = '<unknown>'
    tree = parse(source, filename)
    if args.mode == 'dump':
        print(ast.dump(tree, indent=3, include_attributes=True))
    elif args.mode == 'run':
        compiled = compile(tree, filename, 'exec')
        _run_code(compiled, {
            '__builtins__': builtins
        }, mod_name='__main__', script_name=filename)
    elif args.mode == 'py':
        print(ast.unparse(tree))
    elif args.mode == 'compile_only':
        error: None
        start = time.process_time_ns()
        try:
            compile(tree, filename, 'exec')
        except SyntaxError as e:
            error = e
        end = time.process_time_ns()
        if error is None:
            print(f'Successfully compiled "{filename}" (AST had {count_nodes(tree)} nodes) in {end - start}ns.')
        else:
            print(f'Failed to compile "{filename}" (AST had {count_nodes(tree)} nodes) in {end - start}ns.')
            import traceback
            traceback.print_exception(SyntaxError, error, error.__traceback__, 0)
            return 1


if __name__ == '__main__':
    sys.exit(main())



import ast
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Optional


class TokenType(Enum):
    # Single-character tokens.
    AT = auto()
    CARET = auto()
    COLON = auto()
    LEFT_PAREN = auto()
    RIGHT_PAREN = auto()
    LEFT_BRACE = auto()
    RIGHT_BRACE = auto()
    COMMA = auto()
    DOT = auto()
    MINUS = auto()
    PLUS = auto()
    SEMICOLON = auto()
    TILDE = auto()
    PERCENT = auto()

    # One or two character tokens.
    AMPERSAND = auto()
    AMPERSAND_AMPERSAND = auto()
    PIPE = auto()
    PIPE_PIPE = auto()
    BANG = auto()
    BANG_EQUAL = auto()
    EQUAL = auto()
    EQUAL_EQUAL = auto()
    GREATER = auto()
    GREATER_EQUAL = auto()
    GREATER_GREATER = auto()
    LESS = auto()
    LESS_EQUAL = auto()
    LESS_LESS = auto()
    SLASH = auto()
    SLASH_SLASH = auto()
    STAR = auto()
    STAR_STAR = auto()

    # Literals.
    IDENTIFIER = auto()
    STRING = auto()
    INTEGER = auto()
    DECIMAL = auto()

    # Keywords.
    ASSERT = auto()
    ASYNC = auto()
    AWAIT = auto()
    BREAK = auto()
    CATCH = auto()
    CLASS = auto()
    CONTINUE = auto()
    DEF = auto()
    DEL = auto()
    ELSE = auto()
    FALSE = auto()
    FINALLY = auto()
    FOR = auto()
    FROM = auto()
    GLOBAL = auto()
    IF = auto()
    IMPORT = auto()
    IN = auto()
    IS = auto()
    NONE = auto()
    NONLOCAL = auto()
    NOT = auto()
    RAISE = auto()
    RETURN = auto()
    TRUE = auto()
    TRY = auto()
    WHILE = auto()
    WITH = auto()
    YIELD = auto()

    EOF = auto()


@dataclass(init=True, repr=True)
class Token:
    type: TokenType
    lexeme: str
    line: int
    column: int
    index: int
    literal: Any = None


class TokenGroup:
    SINGLE_COMPARISON = {
        TokenType.BANG_EQUAL,
        TokenType.EQUAL_EQUAL,
        TokenType.LESS,
        TokenType.LESS_EQUAL,
        TokenType.GREATER,
        TokenType.GREATER_EQUAL,
        TokenType.IS,
        TokenType.IN,
    }

    BIT_SHIFT = {
        TokenType.LESS_LESS,
        TokenType.GREATER_GREATER,
    }

    TERMS = {
        TokenType.MINUS,
        TokenType.PLUS,
    }

    FACTORS = {
        TokenType.STAR,
        TokenType.AT,
        TokenType.SLASH,
        TokenType.SLASH_SLASH,
        TokenType.PERCENT,
    }

    UNARY_LOW = {
        TokenType.MINUS,
        TokenType.PLUS,
        TokenType.TILDE,
    }

    LITERALS = {
        TokenType.INTEGER,
        TokenType.DECIMAL,
        TokenType.STRING,
    }


KEYWORDS: dict[str, Optional[TokenType]] = {
    'and':      None,
    'as':       None,
    'assert':   TokenType.ASSERT,
    'async':    TokenType.ASYNC,
    'await':    TokenType.AWAIT,
    'break':    TokenType.BREAK,
    'class':    TokenType.CLASS,
    'continue': TokenType.CONTINUE,
    'def':      TokenType.DEF,
    'del':      TokenType.DEL,
    'elif':     None,
    'else':     TokenType.ELSE,
    'except':   None,
    'false':    TokenType.FALSE,
    'False':    TokenType.FALSE,
    'finally':  TokenType.FINALLY,
    'for':      TokenType.FOR,
    'from':     TokenType.FROM,
    'global':   TokenType.GLOBAL,
    'if':       TokenType.IF,
    'import':   TokenType.IMPORT,
    'in':       TokenType.IN,
    'is':       TokenType.IS,
    'lambda':   None,
    'None':     TokenType.NONE,
    'nonlocal': TokenType.NONLOCAL,
    'not':      TokenType.NOT,
    'or':       None,
    'pass':     None,
    'raise':    TokenType.RAISE,
    'return':   TokenType.RETURN,
    'true':     TokenType.TRUE,
    'True':     TokenType.TRUE,
    'try':      TokenType.TRY,
    'while':    TokenType.WHILE,
    'with':     TokenType.WITH,
    'yield':    TokenType.YIELD,
}

COMPARISON_OPERATORS: dict[TokenType, ast.cmpop] = {
    TokenType.LESS:          ast.Lt,
    TokenType.LESS_EQUAL:    ast.LtE,
    TokenType.GREATER:       ast.Gt,
    TokenType.GREATER_EQUAL: ast.GtE,
    TokenType.EQUAL_EQUAL:   ast.Eq,
    TokenType.BANG_EQUAL:    ast.NotEq,
    # Still technically comparison operators
    TokenType.IN:            ast.In,
    TokenType.IS:            ast.Is,
}

BINARY_OPERATORS: dict[TokenType, ast.operator] = {
    TokenType.LESS_LESS:       ast.LShift,
    TokenType.GREATER_GREATER: ast.RShift,
    TokenType.STAR:            ast.Mult,
    TokenType.AT:              ast.MatMult,
    TokenType.SLASH:           ast.Div,
    TokenType.SLASH_SLASH:     ast.FloorDiv,
    TokenType.PLUS:            ast.Add,
    TokenType.MINUS:           ast.Sub,
    TokenType.PERCENT:         ast.Mod,
}

UNARY_OPERATORS: dict[TokenType, ast.operator] = {
    TokenType.PLUS:  ast.UAdd,
    TokenType.MINUS: ast.USub,
    TokenType.TILDE: ast.Invert,
}



from setuptools import setup


with open('README.md', 'r') as fp:
    long_description = fp.read()


setup(
    name = 'Scython',
    version = '0.1.0',
    description = 'A C-like alternative syntax parser/front-end for Python',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    author = 'Josiah (Gaming32) Glosson',
    author_email = 'gaming32i64@gmail.com',
    url = 'https://github.com/Gaming32/Scython',
    packages = ['scy'],
    license = 'License :: OSI Approved :: MIT License',
    entry_points = {
        'console_scripts': [
            'scy=scy.__main__:main',
        ],
    },
)
